/* 
*   env.c: routines to access environment variables
*   By Terry R. McConnell 1/98
*
*/


#include<stdlib.h>
#include<stdio.h>
#include<string.h>
#include <errno.h>
#include "e2err.h"
#include "env.h"
#include <dos.h>

/* It is assumed the string gotten from the environment has the form
   [devname:][path] where devname may be something like hd3a or fd1, and
   path is a Minix path starting with /. The following routine just splits
   the two components -- it does no validity checking on them. For that,
   use the name2parms routine.

   Returns  value indicating what information was found (cf. env.h)
*/

int 
parse_env_string( char *str, char *dev_name, char *path)
{
	char buffer[1024];
	char *ptr;
	int rval;

	if(str == NULL)return 1;
	if(strlen(str)>1023)return 1;
	if(strlen(str)==0) return NO_DEV_OR_PATH;

	strcpy(buffer,str);

	if((ptr = strchr(buffer,':'))==NULL){    /* No dev given */
		strcpy(dev_name,DEFAULT_DEV);
		strcpy(path,buffer);
		return NO_DEV;
	}

	*ptr = '\0';
	strcpy(dev_name,buffer);
	if(strlen(ptr+1)==0){
		strcpy(path,DEFAULT_PATH);
		return NO_PATH;
	}
	else strcpy(path,ptr+1);
	return PATH_AND_DEV;
}

/* Find the variable from the environment and split its contents using
   parse_env_string. Return 0 if the operation was successful, 1 else.
*/

int
readenv(char *dev_name, char *path)
{
	return parse_env_string(getenv(ENVAR),dev_name,path);
}

static int getenvseg(unsigned *envseg);

/* set_master_env: Set a variable in the environment of the parent process 
   This is highly DOS-dependant */ 

int
set_master_env(char *name, char *value)
{
        unsigned env_bytes = 0;
	unsigned nvars = 0;
        unsigned  master_seg;
        char  *master_ptr;
	char *master_ptr_save;
	char **eptr;
	int i,j,found;

	if(getenvseg(&master_seg)==0){
		printf("Unable to find parent shell environment\n");
		exit(1);
	  } 
	master_ptr_save = master_ptr = (char *)MK_FP(master_seg,0x0000);

        while(strlen(master_ptr)!=0){ 
		if(name == NULL)
                	printf("%s\n",master_ptr);
		env_bytes += (strlen(master_ptr)+1);		
		nvars++;
		master_ptr += (strlen(master_ptr)+1);
	}
	if(name == NULL) return 0;

	/* Build a new copy of the environment so it can be manipulated */
	/* We malloc one extra slot in the array in case a new variable
           is to be added */

	eptr = (char **)malloc((nvars+1)*sizeof(char *));
	if(eptr == NULL)
		e2_err("set_master_env:",ENOMEM);
	master_ptr = master_ptr_save;
	for(i=0;i< nvars;i++){
		eptr[i] = (char *)malloc(strlen(master_ptr)+1);
		if(eptr[i] == NULL) 
			e2_err("set_master_env:",ENOMEM);
			
		strcpy(eptr[i],master_ptr);
		master_ptr += (strlen(master_ptr)+1);
	}

	/* See if variable name exists already */
	found = 0;
	for(i=0;i<nvars;i++)
		if(strncmp(eptr[i],name,strlen(name))==0){
		        if(*(eptr[i]+strlen(name))=='='){	
				found = 1;
				break;
			} /* Otherwise, it just _starts_ the right way */
		}
	/* Here i = index of variable, if found */
	if(found){      /* Existing variable */
		if(value==NULL){  /* remove variable */
			free(eptr[i]);
			for(j=i+1;j<nvars;j++)
				eptr[j-1]=eptr[j];
			nvars--;
		}
		else { 
			eptr[i] = (char *)realloc(eptr[i],strlen(name)+strlen(value)+2);
			if(eptr[i]==NULL)
				e2_err("set_master_env:",ENOMEM);
		  	strcpy(eptr[i],name);
			strcat(eptr[i],"=");
			strcat(eptr[i],value);
		}
	}
	else  {     /* New variable */	
		eptr[nvars] = (char *)malloc(strlen(name)+strlen(value)+2);
		if(eptr[nvars] == NULL)
			e2_err("set_master_env:",ENOMEM);
		strcpy(eptr[nvars],name);
		strcat(eptr[nvars],"=");
		strcat(eptr[nvars++],value);
	}

	/* Now write the altered environment back to the original place */

	master_ptr = master_ptr_save;
	for(i=0;i< nvars;i++){
		strcpy(master_ptr,eptr[i]);
		master_ptr += strlen(eptr[i])+1;
	}

	/* terminate with an extra 0 char */
	*master_ptr = '\0';

	/* clean up */
	for(i=0;i<nvars;i++)free(eptr[i]);
	free(eptr);
	return 0;
}


/* The following code was kindly provided by Mr. Edison Castro in
   response to a question I posted on Comp.os.msdos.programmer
*/

int getenvseg(unsigned *envseg )
{
/*
 * Try to locate the parent environment and fill in the segment address
 * This routine has some OS version dependencies
 * Returns  1  if environment found
 *   0  if environment not found
 */
 unsigned ppsp;
 union REGS  regs;
 int status ;

 regs.x.ax = 0xD44D;      /* 4dos signature value */
 regs.x.bx = 0;
 int86(0x2F, &regs, &regs);    /* TSR multiplex interrput */
 ppsp   = regs.x.cx;      /* psp of immediate parent shell */

 if( regs.x.ax == 0x44dd && ppsp == *(unsigned far *)MK_FP(_psp,0x16) ) {
  /*
   * good, we are running 4dos - AND it is our parent
   * in this case we don't care about dos version
   */
  *envseg =  *(unsigned far *)MK_FP(ppsp,0x2C);
 } else {
  /*
   * Follow the pointers back to our parent psp and from there get the
   * parent environment segment address
   * With early command.com (pre 3.3) the lowest level shell did not fill
   * in the environment segment (it is 0) but we can still find the
   * environment since it always follows the lowest level shell directly
   */
  ppsp = *(unsigned far *)MK_FP(_psp,0x16);
  *envseg =  *(unsigned far *)MK_FP(ppsp,0x2C);
  if ((*envseg == 0) &&
   (_osmajor <= 2 || (_osmajor == 3 && _osminor <= 20))) {
   /*
    * we have the pre 3.3 master command.com - have to assume that the
    * environment segment directly follows the program segment
    * so get the length from the MCB and add it to ppsp
    */
   *envseg = ppsp + *(unsigned far *)MK_FP(ppsp - 1, 3) + 1;
  }
 }
 /*
  * Verify that the owner of the envseg matches the ppsp we got above
  * - just to be a little paranoid
  */
 if ( *(unsigned far *)MK_FP((*envseg)-1, 1) != ppsp) {
  *envseg = 0 ;  /* something wrong :-( */
  status  = 0;
 } else {
  status  = 1 ;
 }
 return(status);
}
