/*  m2ls.c     -- list directories from a Minix file system */
/*  Copyright 1998, Terry R. McConnell    */
/*  Copyright 1988,1991 Steven W. Harrold - All rights reserved. */


#define USAGE "m2ls [-acdfilnrstu1FRS] <[drive][:path]> "
#include    <stdio.h>
#include    <ctype.h>

/* Avoid a conflict with sys/types.h */
#include    <time.h>
#include    <stdlib.h>
#include    <string.h>
#include    <errno.h>
#include    <dos.h>
#include    <io.h>
#include    "diskio.h"
#include    "part.h"
#include    "genhd.h"
#include    "e2err.h"
#include    "dev.h"
#include    "sys/types.h"
#include    "mfs.h"
#include    "env.h"
#include    "passwd.h"

unsigned _stklen = 20000; /* Large: to support recursion */
int direction = 1;  /* sort direction */

/* Types of listings, sorts, etc. Used to set bits in list_type */
#define LS_LONG 1
#define LS_ALL  2
#define LS_F    4
#define LS_U    8
#define LS_D   0x10
#define LS_C   0x20
#define LS_SORTSIZE 0x40
#define LS_SORTTIME 0x80
#define LS_ATIME 0x100
#define LS_CTIME 0x200
#define LS_MTIME 0x400
#define LS_NOSORT 0x800
#define LS_ONECOL 0x1000
#define LS_KB     0x2000
#define LS_I      0x4000
#define LS_R      0x8000

/* function declarations */

struct v2_inode *get_inode(struct devdata *ddata, struct super_block *,
                         ino_t *inum, struct v2_inode *ip, char **path);
void minix_to_dos(char text, struct v2_inode *ip, 
                                     struct devdata *ddata, FILE *outfd);
int parse_env_string( char *str, char *dev_name, char *path);
int readenv(char *, char *);
struct v2_inode *v1_to_v2(struct v1_inode *);
extern read_file(struct directory *darea, struct v2_inode *ip, 
	struct devdata *ddata) ;
void krqsort( void *v[],int left, int right, int (*comp)(void *,void *));
void *binsearch(void *key,void *v[],int n, int (*comp) (void *, void *));
int do_ls(char *minixdir, int list_type, struct devdata *,struct super_block *);


/* The following struct contains all the information in the inode
   we need for various listings
*/

struct fileinfo {
	char      *name;
	mode_t     mode;
	nlink_t    nlinks;
	off_t      size;
	time_t     atime;
	time_t     mtime;
	time_t     ctime;
	ino_t      inum;
	u16_t      uid;
	u16_t      gid;
};


struct user {
	char *name;
	u16_t uid;
};

struct user_group {
	char *name;
	u16_t gid;
};

static struct user *usernames[MAX_NAMES];
static struct user_group *groups[MAX_GROUPS];
int nusers,ngroups;
int numeric;

static char buf[BLOCK_SIZE] ;  /* A buffer for reading disk blocks */

/*===================================================================*/
/* Comparison functions for quick sort    
*/

int
compare_by_name(struct fileinfo *first, struct fileinfo *second)
{
	return direction * strcmp(first->name, second->name);
}

int
compare_by_size(struct fileinfo *first, struct fileinfo *second)
{
	if(first->size == second->size)return 0;
	return direction * (first->size < second->size ? -1: 1);
}

int
compare_by_atime(struct fileinfo *first, struct fileinfo *second)
{
	if(first->atime == second->atime)return 0;
	return direction * (first->atime < second->atime ? -1: 1);
}
int
compare_by_mtime(struct fileinfo *first, struct fileinfo *second)
{
	if(first->mtime == second->mtime)return 0;
	return direction * (first->mtime < second->mtime ? -1: 1);
}

int
compare_by_ctime(struct fileinfo *first, struct fileinfo *second)
{
	if(first->ctime == second->ctime)return 0;
	return direction * (first->ctime < second->ctime ? -1: 1);
}

int
compare_by_uid(struct user *first, struct user *second)
{
	if(first->uid == second->uid)return 0;
	return first->uid < second->uid ? -1 : 1;
}
int
compare_by_gid(struct user_group *first, struct user_group *second)
{
	if(first->gid == second->gid)return 0;
	return first->gid < second->gid ? -1 : 1;
}

/*===================================================================*/

/* perms2str: convert a 16-bit int to a perms string as used in
   ls -l: This is not at all elegant :-( 
*/

int 
perms2str(char *buf, int perms)
{
	int type;

	type = S_IFMT & perms;
	*buf = '\0';

	switch(type){
		case S_IFDIR:
			sprintf(buf,"%s","d");
			break;
		case S_IFCHR:
			sprintf(buf,"%s","c");
			break;
		case S_IFBLK:
			sprintf(buf,"%s","b");
		default:
			sprintf(buf,"%s","-");
			;
	}

	if(S_IRUSR & perms)strcat(buf,"r");
		else strcat(buf,"-");
	if(S_IWUSR & perms)strcat(buf,"w");
		else strcat(buf,"-");
	if(S_ISUID & perms)strcat(buf,"s");
		else
			if(S_IXUSR & perms)strcat(buf,"x");
				else strcat(buf,"-");
	if(S_IRGRP & perms)strcat(buf,"r");
		else strcat(buf,"-");
	if(S_IWGRP & perms)strcat(buf,"w");
		else strcat(buf,"-");
	if(S_ISGID & perms)strcat(buf,"s");
		else
			if(S_IXGRP & perms)strcat(buf,"x");
				else strcat(buf,"-");
	if(S_IROTH & perms)strcat(buf,"r");
		else strcat(buf,"-");
	if(S_IWOTH & perms)strcat(buf,"w");
		else strcat(buf,"-");
	if(S_ISVTX & perms)strcat(buf,"t");
		else
			if(S_IXOTH & perms)strcat(buf,"x");
				else strcat(buf,"-");

	return 0;
}

/*===================================================================*/

/* longlist: print out a line of information in the ls -l format */

void
longlist(struct fileinfo *afile, int listtype)
{
	char time_buf[15];
	char perm_buf[11];
	char name_buf[MAX_USER_NAME_LEN + 1];
	char group_buf[MAX_GROUP_NAME_LEN +1];
	time_t *timep;
	time_t timenow;
	struct tm *time_now;
        int month_now;
        int year_now;
	struct tm *time_then;
	struct user TemplateUser;         /* for searches */
	struct user_group TemplateGroup; 
	struct user *auser;
	struct user_group *agroup;

	/* which time to use ? */
	if(listtype & LS_U)
		timep = (time_t *)&(afile->atime);
	else
	    if(listtype & LS_C )
		timep = (time_t *)&(afile->ctime);
		else timep = (time_t *)&(afile->mtime);

	/* Do we need to give the year rather than the time of day, i.e,
           is the month name ambiguous ? OK if later month in previous
           year, or earlier month in same year */

	time(&timenow);
	time_now = localtime(&timenow);
        month_now = time_now->tm_mon;
        year_now = time_now->tm_year;
	time_then = localtime(timep);

	if(((time_then->tm_year >= (year_now-1))&&
		(time_then->tm_mon > month_now)) ||
	   (time_then->tm_year >= year_now))
        	strftime(time_buf,14,"%b %d %H:%M",localtime(timep));
	else
        	strftime(time_buf,14,"%b %d  %Y",localtime(timep));


	/* convert file mode to rwx etc */
	perms2str(perm_buf, afile->mode);

	/* Print inode number if necessary */
	if(listtype & LS_I)
		printf("%8u ", afile->inum);

	sprintf(name_buf,"%u",afile->uid);
	sprintf(group_buf,"%u",afile->gid);

	if(numeric == 0){
		TemplateUser.uid = afile->uid;
		TemplateGroup.gid = afile->gid; 
		if((auser = (struct user *)binsearch((void *)&TemplateUser, (void **)
                  usernames, nusers, (int (*)(void *,void *))compare_by_uid))
                 != NULL) strcpy(name_buf,auser->name);
		if((agroup = (struct user_group *)binsearch((void *)&TemplateGroup, (void **)
                  groups, ngroups, (int (*)(void *,void *))compare_by_gid))
                 != NULL) strcpy(group_buf,agroup->name);
	}

	if(listtype & LS_KB)
   		printf("%10s %3u %-8s %-8s %9luK %12s %-s", 
			perm_buf,afile->nlinks,name_buf,group_buf,
               		afile->size/1024,time_buf,afile->name);
	else
   		printf("%10s %3u %-8s %-8s %9lu %12s %-s", 
			perm_buf,afile->nlinks,name_buf,group_buf,
               		afile->size,time_buf,afile->name);
	if(listtype & LS_F)
		if(S_ISDIR(afile->mode))
			printf("/");
		else
			if(afile->mode & S_IFEXEC)
				printf("*");
	printf("\n");
	return;
}

/*===================================================================*/

/* Print out successive calls in 3 columns */

void
list_3col(struct fileinfo *afile, int type)
{
	static int count;
	int col;           /* actually, column - 1 */
	char longspaces[] = "          ";
	char shortspaces[] = "  ";
	int i;

	col = count % 3;
	if(type & LS_I){
		printf("%8u ",afile->inum);
		printf("%s",shortspaces);
	}
	else printf("%s",longspaces);
	printf("%-15s",afile->name);
	if(type & LS_F)
		if(S_ISDIR(afile->mode))
			printf("/");
		else
			if(afile->mode & S_IFEXEC)
				printf("*");
	if(col == 2)printf("\n");
	count++;
}



/*==================================================================*/

main (argc, argv)
int     argc ;
char    *argv[] ;
{
    int     drive ;
    int     partno;
    int     err;
    int     list_type = 0;   /* -l -a ... */
    struct devdata      my_ddata;
    struct devdata      *ddata  = &my_ddata;
    long     blkno ;
    struct super_block  sblk ;
    struct part_desc pl[16];
    long offset;
    int     i;
    int     first_time = 1;
    int     more_than_one_dir = 0;
    char    c;
    FILE    *outfd ;
    char    device_name[10];
    char    device_name_save[10];
    char    minixdir[MAX_PATH+NAME_SIZE+2] ;
    char    save_minixdir[MAX_PATH+NAME_SIZE+2];
    struct passwd *pwd;
    struct group *grp;

/*  Does (s)he want help?
 */
    if ((argc >= 2) && (argv[1][0] == '-') && (argv[1][1] == '?'))
    {
printf("Copyright 1998 Terry R. McConnell\n");
printf ("Copyright 1988,1991 Steven W. Harrold - All rights reserved\n") ;
        printf ("Version %s\n", VERSION) ;
        printf ("%s\n",USAGE) ;
        printf ("'drive' is according to Minix convention, e.g. hda3\n") ;
        printf ("'path'is a Minix directory name\n") ;
        exit (0) ;
    }

/* Set up the array of usernames and group names, if the user has supplied
   the passwd and group files */

   if(setpwent()==0){
	i = 0;
	while(((pwd = getpwent()) != NULL)&&(i < MAX_NAMES)){
		usernames[i] = (struct user *)malloc(sizeof(struct user));
		if(usernames[i]==NULL)
			e2_err("m2cp:",ENOMEM);
		usernames[i]->name = (char *)malloc(strlen(pwd->pw_name)+1);
		if(usernames[i]->name ==NULL)
			e2_err("m2cp:",ENOMEM);
		strcpy(usernames[i]->name,pwd->pw_name);
		usernames[i]->uid = pwd->pw_uid;
		i++;
	}
        nusers = i;
     	krqsort((void **)usernames,0,nusers-1, (int (*)(void *,void *))
		compare_by_uid);
	endpwent();
   }

   if(setgrent()==0){
	i = 0;
	while(((grp = getgrent()) != NULL)&&(i < MAX_NAMES)){
		groups[i] = (struct user_group *)malloc(sizeof(struct user_group));
		if(groups[i]==NULL)
			e2_err("m2cp:",ENOMEM);
		groups[i]->name = (char *)malloc(strlen(grp->gr_name)+1);
		if(groups[i]->name ==NULL)
			e2_err("m2cp:",ENOMEM);
		strcpy(groups[i]->name,grp->gr_name);
		groups[i]->gid = grp->gr_gid;
		i++;
	}
        ngroups = i;
     	krqsort((void **)groups,0,ngroups-1, (int (*)(void *,void *))
		compare_by_gid);
	endgrent();
   }
		
		

/* Process command line options  */

    while((argc >= 2) && (argv[1][0] == '-'))
    {
    	i=1;
	while((c = argv[1][i++]) != '\0')
		switch(c){
			case 'a':
				list_type |= LS_ALL;
				break;
			case 'l':
				list_type |= LS_LONG;	
				break;
			case 'd':
				list_type |= LS_D;
				break;
			case 'n':
				numeric = 1;
				break;
			case '1':
				list_type |= LS_ONECOL;
				break;
			case 'i':
				list_type |= LS_I;
				break;
			case 'c':
				list_type |= LS_C;
				list_type &= ~LS_MTIME;
				list_type |= LS_CTIME;
				break;
			case 'u':
				list_type |= LS_U;
				list_type &= ~LS_MTIME;
				list_type |= LS_ATIME;
				break;
			case 'F':
				list_type |= LS_F;
				break;
			case 'r':
				direction = -1;
				break;
			case 'S':
				list_type |= LS_SORTSIZE;
				break;
			case 't':
				list_type |= LS_SORTTIME;
				list_type |= LS_MTIME;
				break;
			case 'f':
				list_type |= LS_NOSORT;
				break;
			case 's':
				list_type |= LS_KB;
				break;
			case 'R':
				list_type |= LS_R;
				break;
			default:
				fprintf(stderr,"Unknown option letter %c\n",c);
				return 1;
		}
        argc-- ;
        argv++ ;
    }
   
	
    device_name[0]='\0';
    device_name_save[0]='\0';
    minixdir[0] = '\0';
    save_minixdir[0] = '\0';

    readenv(device_name_save,save_minixdir); /* See if stuff is given in the environ.
                                      If not, device_name defaults to fd0
                                      and minixdir to /  */

    /* Normalize name of directory by making sure that it ends with / */

    if(*(save_minixdir+strlen(save_minixdir)-1)!='/')
	strcat(save_minixdir,"/");

    if(argc > 2)more_than_one_dir = 1;

    while(1) { /* loop until no further command line args */

       if(argc == 1) /* Can only happen when no directories given */
	  i = NO_DEV_OR_PATH;
       else
          i = parse_env_string(argv[1],device_name,minixdir);
		
       if((first_time-- > 0)||((i!=NO_DEV)&&(i!=NO_DEV_OR_PATH))){

	      if((i==NO_DEV)||(i==NO_DEV_OR_PATH))
		strcpy(device_name,device_name_save);

	      /* After first time through: we'll assume if they specified
                 device it's because they are changing device */

              /* Read partition tables and fill descriptor array. Convert Device
              identifier name to drive number and partition index */

    		name2parms(&drive,&partno,device_name,pl);

    		/* Load ddata structure */
    		init_fs(drive,partno,pl,&my_ddata);


		/*  Read the super block */

    		offset = ddata->pl[ddata->d_partnum].start; /* offset in sectors of fs
                                                    on disk */
    		blkno = 2*SUPER_BLOCK + offset;

    		err = readdisk (ddata->d_drive, blkno, 1,buf) ;
    		if(err)
        		e2_err("m2ls:readdisk",err);

    		memcpy (&sblk, buf, sizeof(sblk)) ;

    		if (((sblk.s_magic != MINIX_V2_SUPER_MAGIC) && (sblk.s_magic
        		!= MINIX_V1_SUPER_MAGIC))        ||
        		(sblk.s_ninodes < 1)            ||
        		(sblk.s_imap_blocks < 1)        ||
        		(sblk.s_zmap_blocks < 1))
        		e2_err("m2cp",M2E_BADFS);

		/*  Store zone to block conversion in dev data struct */
		ddata->d_log_zone_size = sblk.s_log_zone_size;

	}

       if((i!=NO_PATH)&&(i!=NO_DEV_OR_PATH)){
       		if(minixdir[0]!='/') { /* relative path */ 
                        i = strlen(minixdir);
    	   		strcat(save_minixdir,minixdir);
			strcpy(minixdir,save_minixdir);
			save_minixdir[i]='\0';
        	}
	}
	else strcpy(minixdir,save_minixdir);

    /* Just do it */

    if(more_than_one_dir)
	printf("\n%s:%s:\n\n",device_name,minixdir);

    do_ls(minixdir, list_type, ddata, &sblk);

    if(--argc <= 1)break;
    argv++;
   }

    /* NB: We have not freed the arrays of users and groups */
    return 0;

} /* main() */

/* do_ls: This does the bulk of the work. It needs to be separated out
   so that it can be called recursively on directories if need be */

int
do_ls(char *minixdir, int list_type, struct devdata *ddata, struct super_block *sblk)
{

    struct v2_inode  *dip, *rip, *tip ;
    struct directory    *darea , *dp;
    struct fileinfo    **farea;
    int dnz,dn,dn_actual;
    int     inum;
    char    buffer[MAX_PATH];
    char    *path[MAX_PATH/2+1] ;
    char    *newpath;
    char    *tp ;
    int i;

    strcpy(buffer,minixdir);

/*  Break Minix filename into path components */

    if (buffer[0] != '/')
    {
        fprintf (stderr,
                "Final path %s is not absolute\n",buffer) ;
        return 1 ;
    }

    tp = strtok (buffer, "/") ;
    i = 0 ;
    while (tp)
    {
        path[i++] = tp ;
        tp = strtok (NULL, "/") ;
    }
    path[i] = NULL ;
     inum = 1;
     rip = get_inode(ddata,sblk,(ino_t *)&inum,NULL,NULL); /* 1 is root directory inum */

/*  rip now points to a copy of the root directory inode in memory. We
    pass this to get_inode, which searches for the inode of the desired
    directory. If successful, this will return a pointer to an in-memory copy
    of the desired inode.
 */

  if(strcmp(minixdir,"/")!=0){
    dip = get_inode(ddata,sblk, (ino_t *)&inum, rip, path) ;    /* recursively scan */

    if (!dip)
    {
        fprintf (stderr, "Cannot locate Minix directory '%s'\n", minixdir) ;
        exit (1) ;
    }
    if (!(dip->i_mode & S_IFDIR))
    {
        fprintf (stderr, "Minix file '%s' is not a directory\n",
                         minixdir) ;
        exit(1) ;
    }
    free(rip); /* Not used further */
  }
  else dip = rip;

    /* We now read the directory with inode dip into memory */

    dn = dip->i_size / sizeof(struct directory) ;
    dnz = dn * sizeof(struct directory) + BLOCK_SIZE - 1 ;
    dnz /= BLOCK_SIZE ;
    dnz *= BLOCK_SIZE ;
    dnz /= sizeof(struct directory) ;
    darea = calloc (dnz+1, sizeof(struct directory)) ;
    if (!darea)
	e2_err("m2ls:do_ls",ENOMEM);

    read_file(darea, dip, ddata) ;

    /* Setup an array of pointers to fileinfo entries. This
       is necessary in order to be able to sort the directory entries
       on various fields. Also, the directory may contain many "phantom"
       files -- ones which have been unlinked but whose names have not
       been removed from the directory. */

    farea = (struct fileinfo **)malloc(dn*sizeof(struct fileinfo *));
    if(farea==NULL)
	e2_err("m2ls:do_ls:",ENOMEM);

    if( list_type & LS_D ){
	dn_actual = 1;
	farea[0] = (struct fileinfo *)malloc(sizeof(struct fileinfo));
	if(farea[0]==NULL)
		e2_err("m2ls:do_ls",ENOMEM);
	i = 0;
	while(1){
                farea[0]->name = malloc(NAME_SIZE+1);
		if(farea[0]->name == NULL)
			e2_err("do_ls",ENOMEM);
		strcpy(farea[0]->name, path[i]);
		if(path[++i] == NULL)break;
	} 
	farea[0]->mode = dip->i_mode;	
	farea[0]->nlinks = dip->i_nlinks;
	farea[0]->uid = dip->i_uid;
	farea[0]->gid = dip->i_gid;
	farea[0]->size = dip->i_size;
	farea[0]->inum = inum;
    }
    else {
        dp = darea;
        dn_actual = 0;
    	for(i=0;i<dn; i++,dp++){
		if(dp->d_inum){
      			tip = get_inode(ddata,sblk,(ino_t *)&(dp->d_inum),NULL,NULL);
			farea[dn_actual] = (struct fileinfo *)malloc(sizeof(struct fileinfo));
			if(farea[dn_actual]==NULL)
				e2_err("m2ls:do_ls",ENOMEM);
                        farea[dn_actual]->name = calloc(1,NAME_SIZE+1);
			if(farea[dn_actual]->name == NULL)
				e2_err("do_ls:",ENOMEM);
			strncpy(farea[dn_actual]->name,dp->d_name,NAME_SIZE);
			farea[dn_actual]->mode = tip->i_mode;	
			farea[dn_actual]->nlinks = tip->i_nlinks;
			farea[dn_actual]->uid = tip->i_uid;
			farea[dn_actual]->gid = tip->i_gid;
			farea[dn_actual]->size = tip->i_size;
			farea[dn_actual]->atime = tip->i_atime;
			farea[dn_actual]->mtime = tip->i_mtime;
			farea[dn_actual]->ctime = tip->i_ctime;
			farea[dn_actual]->inum = dp->d_inum;
			dn_actual++;
			free(tip);
		}
    	}
	
/* Sort appropriately */

   	if(!(list_type & LS_NOSORT)){
    		if(list_type & LS_SORTSIZE)
     			krqsort((void **)farea,0,dn_actual-1,
				(int (*)(void *, void *))compare_by_size);
    		else if(list_type & LS_SORTTIME) {
           		if(list_type & LS_ATIME)
             			krqsort((void **)farea,0,dn_actual-1,
	      				(int (*)(void *, void *))compare_by_atime);
           		else if(list_type & LS_CTIME)
             			krqsort((void **)farea,0,dn_actual-1,
	      				(int (*)(void *, void *))compare_by_ctime);
           			else if(list_type & LS_MTIME)
             				krqsort((void **)farea,0,dn_actual-1,
	      					(int (*)(void *, void *))compare_by_mtime);
    		}
    		else
     			krqsort((void **)farea,0,dn_actual-1, 
				(int (*)(void *,void *))compare_by_name);
   	}
   }

/* Print out stuff */

    if(list_type & LS_LONG)
    	printf("total %d\n",dn_actual);
    for (i=0; i<dn_actual; i++)
     if((list_type & LS_R)&&S_ISDIR(farea[i]->mode)&&(strcmp(farea[i]->name,
	".")!= 0)&&(strcmp(farea[i]->name,"..")!=0))
     {
	/* Call do_ls recursively on the present directory */
	newpath = (char *)malloc(MAX_PATH+1);
	if(newpath == NULL)
		e2_err("m2ls:do_ls:", ENOMEM);
	strcpy(newpath,minixdir);
	if(minixdir[strlen(minixdir)-1] != '/')
		strcat(newpath,"/");
	strcat(newpath,farea[i]->name);
	printf("\n%s:\n",newpath);
	do_ls(newpath,list_type,ddata,sblk);
	free(newpath);
     }
     else
      if((list_type & LS_ALL)||((farea[i]->name)[0] != '.')) 
	if( list_type & LS_LONG)
		longlist(farea[i],list_type);
	else
  		if( !isatty(1) || (list_type & LS_ONECOL )){
			if(list_type & LS_I)
				printf("%8u ",farea[i]->inum);
			printf("%s",farea[i]->name);
			if(list_type & LS_F)
				if(S_ISDIR(farea[i]->mode))
					printf("/");
				else
					if(farea[i]->mode & S_IFEXEC)
						printf("*");
			printf("\n");
		}
	else
		list_3col(farea[i],list_type);

	printf("\n");
    

/*  Clean up
 */
    free(darea);
    free(dip);
    for(i=0;i<dn_actual;i++){
	free(farea[i]->name);
	free(farea[i]);
    }
    free(farea);
    return 0;

}
/* This is the implementation of C.A.R. Hoare's quicksort algorthm
	taken from K&R. 

usage: if array[] is an array of some kind of thingies,
       and comp(thingie *a, thingie *b) returns -1, 0, or 1 according
       as a < b, a==b, or a > b in some ordering it imposes on
       thingies, then

       krqsort( array, int left, int right, &comp ) 

       sorts array in order, using initial choices left and right
       in the algorithm.

	Declare this guy in your program as

	void krqsort ( void *arrayname[], int, int,
	 int (*functionname)(void *, void *));

	see sortargs.c for an example using this function

*/

void krqsort( void *v[],int left, int right, int (*comp)(void *,void *))
/*                                                ^
                                                  |
                                            pointer to a function of
                                            two generic pointers,
                                            returning int. */
{
	int i,last;
	void swap(void *v[], int, int );  /* private swap function */

	if( left >= right ) /* do nothing if array contains fewer
                               than 2 elements */	
	return;
	swap(v, left, (left + right)/2);
	last = left;
	for ( i= left +1;i <= right; i++)
		if((*comp)(v[i],v[left]) < 0)
			swap(v,++last,i);
	swap(v,left,last);
	krqsort(v, left, last -1,comp);
	krqsort(v,last+1,right,comp);
}

void swap(void *v[],int i, int j)
{
	void *temp;

	temp = v[i];
	v[i] = v[j];
	v[j] = temp;
}


void *binsearch(void *key,void *v[],int n, int (*comp) (void *, void *))
{
        int cond;
        int low,high,mid;

        low = 0;
        high = n-1;
        while(low <= high){
                mid = (low + high)/2;
                if((cond = (*comp)(key,v[mid])) < 0)
                        high = mid -1;
                else if (cond > 0)
                        low = mid + 1;
                else return v[mid];
        }
		  return NULL;
}

#if 0

/* Debugging routine */
int
dump_inode(struct v2_inode *ip){

	int i;
	printf("mode = %u\n",ip->i_mode);
	printf("i_nlinks = %c\n",'0'+ip->nlinks);
	printf("size = %lu\n",ip->i_size);
	for(i=0;i<10;i++)
		printf("zone %d = %lu\n",i,ip->i_zone[i]);
	return 0;
}
#endif


/*---eof---*/
