/* m2super: print information on superblocks of minix filesystems */
/* By Terry R. McConnell 1998 */

#define USAGE "Usage: m2super [ -a -? <device> ]"

#define INFO "\n\
M2super.exe: print information on superblocks of minix filesystems.\n\n\
By Terry R. McConnell (trmcconn@syr.edu) 1/1998\n\
\n\n\
Usage: m2super [ -a -? <device:> ]\n\
where device is a device name in Minix format (e.g, hd3a).\n\
-a: print info on all Minix filesystems on first two ATA fixed disks.\n\
If no arguments are given, gets device name from the M2CWD envar.\n"

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include "sys/types.h"
#include "mfs.h"
#include "e2err.h"
#include "part.h"
#include "diskio.h"
#include "dev.h"
#include "env.h"

extern build_device_name( int, int, char *, struct part_desc *);
extern int readenv(char *, char *);
extern int parse_env_string(char *, char *, char *);
void print_sb(struct super_block *);

int
main(int argc, char **argv)
{

	int disk,partno,fs_vers,i;
	int flag_all = 0;
	int err;
	struct part_desc pl[16];
	unsigned long offset;
	unsigned short my_sector[256];  /* 512 bytes */
	char device_name[10];
	char path[256];
	struct super_block *my_sb;

/* Process command line */

	if(argc > 2) {
		fprintf(stderr,"%s\n",USAGE);
		return 1;
	}	

	if(argc == 2){ 
		if(strcmp(argv[1],"-?")==0){
			fprintf(stderr,"%s",INFO);
			return 0;
		}
		if(strcmp(argv[1],"-a")==0)flag_all = 1;
		else
			if(parse_env_string(argv[1],device_name,path)==NO_DEV){
				fprintf(stderr,"Invalid device name on command line\n");
				return 1;
			}
	}
	else  /* No arguments */
        	if(readenv(device_name,path)==1){
			fprintf(stderr,"Device name must be in environment or on command line\n");
			return 1;
		}
     if(!flag_all){
	name2parms(&disk,&partno,device_name,pl);
	fs_vers = getfstype(disk,pl+partno);
	if(fs_vers < 0)
		e2_err("m2super:getfstype", -fs_vers);

	if((fs_vers != MINIXV1FS)&&(fs_vers != MINIXV2FS))
		e2_err("m2super",M2E_BADFS);

	if(fs_vers == MINIXV1FS)
	   printf("Minix V1 Filesystem %s (disk  0x%x, partition index %d)\n\n",device_name,disk,partno+1);
	else
	   printf("Minix V2 Filesystem %s (disk  0x%x, partition index %d)\n\n",device_name,disk,partno+1);

	/* Read super-block */	

	offset = pl[partno].start;
	printf("Starting sector of filesystem = %lu\n",offset);
	printf("Partition size = %lu bytes = %lu blocks = %lu sectors\n\n",
		pl[partno].length*SECTOR_SIZE,pl[partno].length*SECTOR_SIZE/BLOCK_SIZE,
			pl[partno].length);

	err=readdisk(disk, offset+2, 1, &my_sector);
	if (err) e2_err("m2super:readdisk", -err);

	my_sb = (struct super_block *)&my_sector;
	print_sb( my_sb);

     }
	else {   /*  -a option */
		/* First try disk 0x80 */
	     for(disk = 0x80;disk<0x82;disk++) {
		err = get_part(disk,pl);
		if(err)
			e2_err("m2super:get_part",err);
		for(i=0;i<16;i++)
			if(pl[i].parttype == PART_MINIX){
				build_device_name(disk,i,device_name,pl);
				fs_vers = getfstype(disk,pl+i);
				if(fs_vers < 0)
					e2_err("m2super:getfstype", -fs_vers);

				if((fs_vers != MINIXV1FS)&&(fs_vers != MINIXV2FS))
					continue;

				if(fs_vers == MINIXV1FS)
	   				printf("Minix V1 Filesystem %s (disk  0x%x, partition index %d)\n\n",device_name,disk,i+1);
				else
	   				printf("Minix V2 Filesystem %s (disk  0x%x, partition index %d)\n\n",device_name,disk,i+1);

				offset = pl[i].start;

					printf("Starting sector of filesystem = %lu\n",offset);
					printf("Partition size = %lu bytes = %lu blocks = %lu sectors\n\n",
						pl[i].length*SECTOR_SIZE,pl[i].length*SECTOR_SIZE/BLOCK_SIZE,
							pl[i].length);
				err=readdisk(disk, offset+2, 1, &my_sector);
				if (err) e2_err("m2super:readdisk", -err);

				my_sb = (struct super_block *)&my_sector;
				print_sb( my_sb);
				printf("------------------\n\n");
			}
              }
     }
     return 0;
}


void
print_sb( struct super_block *my_sb)
{

	/* Print out informative stuff */

	printf("Relevant fields in super-block (decimal,unless noted):\n");
	printf("%-16s = %d\t Usable number of inodes on minor device\n",
		"s_ninodes",my_sb->s_ninodes);
	printf("%-16s = %u\t Total device size, including bitmaps etc\n",
		"s_nzones",my_sb->s_nzones);
	printf("%-16s = %d\t # of blocks used by inode bit map\n",
		"s_imap_blocks",my_sb->s_imap_blocks);
	printf("%-16s = %d\t # of blocks used by zone bit map\n",
		"s_zmap_blocks",my_sb->s_zmap_blocks);
	printf("%-16s = %d\t # of first data zone\n",
		"s_firstdatazone",my_sb->s_firstdatazone);
	printf("%-16s = %d\t log2 of blocks/zone \n",
		"s_log_zone_size",my_sb->s_log_zone_size);
	printf("%-16s = %lu\t maximum file size on this device \n",
		"s_max_size",my_sb->s_max_size);
	printf("%-16s = %X\t (Hex) magic number to recognize super-blocks \n",
		"s_magic",my_sb->s_magic);
	printf("%-16s = %lu\t number of zones (replaces s_nzones in V2) \n",
		"s_zones",my_sb->s_zones);
	return;
}
