/*  mgetinod.c  -- get requested inode entry */
/*  Copyright 1998 Terry R. McConnell */
/*  Copyright 1988,1991 Steven W. Harrold - All rights reserved.*/
/*  $Header: MGETINOD.C_V 1.2 91/03/19 09:57:56 SWH Exp $ */

#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>
#include    <errno.h>
#include    "diskio.h"
#include    "part.h"
#include    "sys/types.h"
#include    "mfs.h"
#include    "dev.h"
#include    "e2err.h"

#define V2_INODES_PER_BLOCK  (BLOCK_SIZE/sizeof(struct v2_inode))
#define V1_INODES_PER_BLOCK  (BLOCK_SIZE/sizeof(struct v1_inode))

static char Template_Block[BLOCK_SIZE];

struct v2_inode* v1_to_v2(struct v1_inode *mynode);
void read_file(void *, struct v2_inode *, struct devdata *);

/*==================================================================*/
/* get_inode: If path is NULL, finds the inode with given inum on disk.
   (In this case, the dip arg is not used ) Otherwise, finds the
   inode for the file with name given by path. The search begins in the
   directory with in-memory inode pointed to by dip. (Normally the root
   directory.) In either case, memory is
   allocated and a pointer to a copy of the desired inode in
   memory is returned. 
   The inode is always of V2 type: if the filesystem is V1 then
   the inode found is converted to V2 format.
*/

struct v2_inode *get_inode(struct devdata *ddata, struct super_block *sblk,
 ino_t *inum, struct v2_inode *dip, char **path)
{
    struct directory    *darea, *dp ;
    struct v1_inode *temp;
    struct v2_inode *ip;
    int                 dn, dnz ;
    ino_t finum;
    int inode_startblock;
    int         i, j, k, n ;
    int blkno;
    char        c, *p ;
    char        npath[MAX_PATH+1] ;
    BOOL        found ;
    unsigned long offset;

    inode_startblock = 1 + sblk->s_imap_blocks + sblk->s_zmap_blocks
                           + SUPER_BLOCK;

  if(path){
    dn = dip->i_size / sizeof(struct directory) ;
    dnz = dn * sizeof(struct directory) + BLOCK_SIZE - 1 ;
    dnz /= BLOCK_SIZE ;
    dnz *= BLOCK_SIZE ;
    dnz /= sizeof(struct directory) ;
    darea = calloc (dnz+1, sizeof(struct directory)) ;
    if (!darea)
	e2_err("get_inode",ENOMEM);

    read_file(darea, dip, ddata) ;

/*  Scan the directory for the name of the current path component
    Note that entries that have been removed from the directory are
    marked with an inode number of 0.
 */
    dp = darea ;
    found = FALSE ;
    for (i=0; !found && i<dn; i++, dp++){
        if (dp->d_inum && (strncmp (dp->d_name, *path, NAME_SIZE) == 0))
        {
            found = TRUE ;
            break ;
        }
    }
    

    if (!found)
    {
        ip = NULL ;
        goto quit ;
    }
    finum = dp->d_inum;
  }
  else finum = *inum;

/* offset in sectors of fs from start of disk */
    offset = ddata->pl[ddata->d_partnum].start;

    if(ddata->d_fs_vers == MINIXV2FS) 
    {
    	/* Reserve memory for next inode */
    	ip = (struct v2_inode *)calloc(1,sizeof(struct v2_inode));

    	/* Block containing target inode 
	relative to beginning of fs */
    	blkno =  finum % V2_INODES_PER_BLOCK ?
	 finum/V2_INODES_PER_BLOCK + inode_startblock :
	   finum/V2_INODES_PER_BLOCK + inode_startblock -1 ; 

   	/* Read the block containing the target inode from disk */
    	readdisk(ddata->d_drive,2*blkno + offset,2,Template_Block);
    
        /* Offset of the desired inode within the block */
    	i = finum % V2_INODES_PER_BLOCK ? 
		finum % V2_INODES_PER_BLOCK - 1 :
		finum % V2_INODES_PER_BLOCK + V2_INODES_PER_BLOCK - 1 ;

    	memcpy(ip,Template_Block+i*sizeof(struct v2_inode),sizeof(struct v2_inode));
     }
     else {
    	temp = (struct v1_inode *)calloc(1,sizeof(struct v1_inode));
    	blkno =  finum % V1_INODES_PER_BLOCK ?
	 finum/V1_INODES_PER_BLOCK + inode_startblock :
	   finum/V1_INODES_PER_BLOCK + inode_startblock -1 ; 
    	readdisk(ddata->d_drive,2*blkno + offset,2,Template_Block);
    	i = finum % V1_INODES_PER_BLOCK ? 
		finum % V1_INODES_PER_BLOCK - 1 :
		finum % V1_INODES_PER_BLOCK + V1_INODES_PER_BLOCK - 1 ;
    	memcpy(temp,Template_Block+i*sizeof(struct v1_inode),sizeof(struct v1_inode));
	ip = v1_to_v2(temp);
	free(temp);
     }	

	/*  Now, for those entries that are themselves directories,
 	*  search them recursively.
 */
    if(!path) goto quit;
    path++ ;
    if (!*path)
    {
        goto quit ;
    }

    *inum = finum;
    ip = get_inode(ddata, sblk,(ino_t *)&inum, ip, path) ;

/*  Clean up
 */
quit:
    if(path)
        free (darea) ;
    return ip ;

} /* get_inode() */

/* Takes a pointer to a version 1 inode. Returns a pointer to a version
   2 inode. This is so we don't have to write 2 versions of everything. */

struct v2_inode* v1_to_v2(struct v1_inode *mynode)
{
	
	struct v2_inode *i2ptr;
	int i;

    	i2ptr = (struct v2_inode *)calloc(1,sizeof(struct v2_inode));
	if(!i2ptr)
		e2_err("v1_to_v2",ENOMEM);

	i2ptr->i_mode = (int) mynode->i_mode;
	i2ptr->i_nlinks = (unsigned short)mynode->i_nlinks;
	i2ptr->i_uid = (int) mynode->i_uid;
	i2ptr->i_gid = (unsigned short)mynode->i_gid;
	i2ptr->i_size = mynode->i_size;
	i2ptr->i_atime = mynode->i_modtime; /* Well, what else can we do ? */
	i2ptr->i_mtime = mynode->i_modtime;
	i2ptr->i_ctime = mynode->i_modtime;
	for(i=0;i<V2_NR_TZONES-3;i++)
		i2ptr->i_zone[i] = (ulong)mynode->i_zone[i];

	i2ptr->i_ind = (ulong)mynode->i_zone[i++];
	i2ptr->i_dbl_ind = (ulong)mynode->i_zone[i];
	return i2ptr;
}
