/***************************************************************************
 * part.c - Routines for handling disk partitions
 *
 * Copyright (C) 1998 Terry McConnell trmcconn@syr.edu
 *
 * Copyright (C) 1995 Claus Tondering, ct@login.dknet.dk
 * This file may be redistributed under the terms of the GNU Public License.
 ***************************************************************************/
 
#include <string.h>
#include <sys/types.h>
#include "genhd.h"
#include "part.h"
#include "diskio.h"
#include "e2err.h"
#include "dev.h"
#include "sys/types.h"
#include "mfs.h"


struct firstsector fs, fs2;

#if 1
void dump_sector(struct firstsector *);
#endif

/****************************************************************************
 *
 * get_part(): Fills array of up to 16 partition descriptors from the disk
 * with given BIOS disk number.
 * 
 * To keep things simple we fill the table in a straightforward order: first
 * we create an entry for each primary partition. Then for each primary
 * Minix partition we fill additional entries for each logical subpartition.
 *
 * Thus, for example, a disk with 4 primary partitions and two logical
 * partitions on each of hd1 and hd3 will have partition descriptors created
 * in the following order: hd1 hd2 hd3 hd4 hd1a hd1b hd3a hd3b. We do not
 * expand extended partitions (although code for doing so is commented out
 * below -- Minix discourages putting minix partitions inside extended
 * partitions.) 
 *
 * The index of a partition can be recovered from the name (e.g. hd3a) by 
 * using the name2parms routine, which does a search through the partition
 * descriptor array.
 *
 * Returns an error code
 ****************************************************************************/

int
get_part(int disk, struct part_desc *pl)
{
	int i,j, next, err,fstype;
	unsigned long startsect, offset,len;

	memset(pl,0,sizeof(*pl)*16);
	if (!(disk & 0x80)) {
		/* Floppy disk */
		pl[0].start = 0L;
		pl[0].length = 2*1440; /* This is, of course, rather
                                       presumptive. It isn't used for
                                       anything currently.
                                        */
		pl[0].parent = -1;
		pl[0].parttype = PART_MINIX; /* Innocent till proven guilty */
		fstype=getfstype(disk, pl);
		if ((fstype==MINIXV2FS) || (fstype == MINIXV1FS))
			pl[0].parttype = PART_MINIX;
		else pl[0].parttype = 0;
		return 0;
	}
	
#if 0
	printf("Reading sector 0\n");
#endif

	err=readdisk(disk, 0, 1, &fs);
	if (err) return err;

#if 0
	dump_sector (&fs);
#endif

	if (fs.magic!=FIRST_SECT_MAGIC)
		return M2E_BADPART;
	
	for (i=0; i<4; i++) {
		if (fs.part[i].sys_ind) {
			pl[i].is_extended = fs.part[i].sys_ind==EXTENDED_PARTITION;
			pl[i].start = fs.part[i].start_sectlo+
			  (((unsigned long)fs.part[i].start_secthi)<<16);
			pl[i].length = fs.part[i].nr_sectslo+
			  (((unsigned long)fs.part[i].nr_sectshi)<<16);
			pl[i].parttype = fs.part[i].sys_ind;
			pl[i].parent = -1;
		}
	}
	next=4;

/* Handle any Minix logical partitions */

	for (i=0; i<4; i++) {
		if (fs.part[i].sys_ind==PART_MINIX) {
			startsect = pl[i].start;

			err=readdisk(disk, startsect, 1, &fs2);
			if (err) return err;
				
			if (fs2.magic!=FIRST_SECT_MAGIC)
				continue;  /* Next i: no subpartitions */
			for(j=0;j<4;j++){
			    len = fs2.part[j].nr_sectslo+
				  (((unsigned long)fs2.part[j].nr_sectshi)<<16);
			    if(len){
				pl[next].start =  fs2.part[j].start_sectlo+
				  (((unsigned long)fs2.part[j].start_secthi)<<16);
				pl[next].length = len;
				pl[next].parttype = fs2.part[j].sys_ind;
				pl[next].parent = i;
				next++;
			     }
			}
		}
#if 0
	/* Unused code to handle extended partitions */
		if (fs.part[i].sys_ind==EXTENDED_PARTITION) {
			startsect = pl[i].start;
			offset = 0;
			while (1) {
				if (next==16)
					return 0;

				err=readdisk(disk, startsect+offset, 1, &fs2);
				if (err) return err;
				
				if (fs2.magic!=FIRST_SECT_MAGIC)
					return E2E_BADPART;

				pl[next].start = startsect + offset + fs2.part[0].start_sectlo+(fs2.part[0].start_secthi<<16);
				pl[next].length = fs2.part[0].nr_sectslo+(fs2.part[0].nr_sectshi<<16);
				next++;
				if (fs2.part[1].sys_ind!=EXTENDED_PARTITION)
					break;
				offset = fs2.part[1].start_sectlo+(fs2.part[1].start_secthi<<16);
			}
		}
#endif
	}
	return 0;
}


/**********************************************************************
 * getfstype finds the type of the file system on the partition.
 * It is not very good at it, but it identifies minix file systems,
 * which is the important thing.
 *
 * Returns the file system type or the negative of an error code.
 **********************************************************************/

int
getfstype(int disk, struct part_desc *pl)
{
	unsigned parttype;
	unsigned long start;


	static union {
		u_char	bytes[512];
		u_short shorts[256];
		u_long	longs[128];
	} testblock;

		
	int err;

	parttype = pl[0].parttype;
	start = pl[0].start;

	if (parttype == PART_SWAP)
		return SWAP;

	if (parttype == PART_DOS32)
		return MSDOSFS;



#if 0
	err=readdisk(disk, start, 1, &testblock);
	if (err) return -err;
	if (strncmp(testblock.bytes+3,"MSDOS",5)==0) /* There must be a better way to do this */
		return MSDOSFS;
#endif

	err=readdisk(disk, start+2, 1, &testblock);
	if (err) return -err;

	if(parttype == PART_MINIX){ 
		if(testblock.shorts[8]==MINIX_V2_SUPER_MAGIC)
			return MINIXV2FS;
		if(testblock.shorts[8]==MINIX_V1_SUPER_MAGIC)
			return MINIXV1FS;
	} 

	err=readdisk(disk, start+7, 1, &testblock);
	if (err) return -err;

	if (memcmp(testblock.bytes+0x1f6, "SWAP-SPACE", 10)==0)
		return SWAP;

	return 0;
}


/* This is just an interface with the the old mintools code */

int
init_fs(int disk, int partno, struct part_desc *pl, struct devdata *dev)
{

	int fs_vers,i;
	int err;

	fs_vers = getfstype(disk,pl+partno);
	if(fs_vers < 0)
		e2_err("init_fs:getfstype", -fs_vers);

	if((fs_vers != MINIXV1FS)&&(fs_vers != MINIXV2FS))
		e2_err("init_fs",M2E_BADFS);

	/* fill dev_data from appropriate fields in super_block */
	
	dev->d_drive = disk;
	dev->d_partnum = partno;
	dev->pl = pl;   /* New field */
	dev->d_fs_vers = fs_vers;
	return 0;
}


#if 0
void dump_sector(struct firstsector *f)
{
	int i;

	printf("Initial 0x1be bytes:");
	for (i=0; i<0x1be; i++) {
		if ((i&0xf)==0) printf("\n");
		printf("%02x ",f->fill[i]);
	}
	
	for (i=0; i<4; i++) {
		printf("\nPartition descriptor %i:\n",i);
		printf("    boot_ind: %02x\n",f->part[i].boot_ind);
		printf("    head: %02x\n",f->part[i].head);
		printf("    sector: %02x\n",f->part[i].sector);
		printf("    cyl: %02x\n",f->part[i].cyl);
		printf("    sys_ind: %02x\n",f->part[i].sys_ind);
		printf("    end_head: %02x\n",f->part[i].end_head);
		printf("    end_sector: %02x\n",f->part[i].end_sector);
		printf("    end_cyl: %02x\n",f->part[i].end_cyl);
		printf("    start_sectlo: %04x\n",f->part[i].start_sectlo);
		printf("    start_secthi: %04x\n",f->part[i].start_secthi);
		printf("    nr_sectslo: %04x\n",f->part[i].nr_sectslo);
		printf("    nr_sectshi: %04x\n",f->part[i].nr_sectshi);
	}

	printf("Magic: %04x\n",f->magic);
	printf("----------------------------------------\n");
}
#endif
